@@ -9,6 +9,7 @@
 #include "session.h"
 #include "settings.h"
 #include "datastructures.h"
+#include "utils.h"
 
 #if GTK_MAJOR_VERSION == 2
 #include "gtk2-compat.h"
@@ -66,25 +67,6 @@ completion_element_free(girara_completion_element_t* element)
   g_slice_free(girara_completion_element_t,  element);
 }
 
-static char*
-escape(const char* value)
-{
-  if (value == NULL) {
-    return NULL;
-  }
-
-  GString* str = g_string_new("");
-  while (*value != '\0') {
-    const char c = *value++;
-    if (strchr("\\ \t\"\'", c) != NULL) {
-      g_string_append_c(str, '\\');
-    }
-    g_string_append_c(str, c);
-  }
-
-  return g_string_free(str, FALSE);
-}
-
 girara_completion_t*
 girara_completion_init()
 {
@@ -261,7 +243,11 @@ girara_isc_completion(girara_session_t* session, girara_argument_t* argument, gi
    *  there is no current list
    */
   if (session->gtk.results == NULL) {
+#if GTK_MAJOR_VERSION == 2
     session->gtk.results = GTK_BOX(gtk_vbox_new(FALSE, 0));
+#else
+    session->gtk.results = GTK_BOX(gtk_box_new(GTK_ORIENTATION_VERTICAL, 0));
+#endif
 
     if (session->gtk.results == NULL) {
       g_free(current_command);
@@ -468,7 +454,7 @@ girara_isc_completion(girara_session_t* session, girara_argument_t* argument, gi
 
     /* update text */
     char* temp;
-    char* escaped_value = escape(((girara_internal_completion_entry_t *) entries_current->data)->value);
+    char* escaped_value = girara_escape_string(((girara_internal_completion_entry_t *) entries_current->data)->value);
     if (command_mode == true) {
       char* space = (n_elements == 1) ? " " : "";
       temp = g_strconcat(":", escaped_value, space, NULL);
@@ -500,14 +486,19 @@ girara_isc_completion(girara_session_t* session, girara_argument_t* argument, gi
 static GtkEventBox*
 girara_completion_row_create(girara_session_t* session, const char* command, const char* description, bool group)
 {
-  GtkBox      *col = GTK_BOX(gtk_hbox_new(FALSE, 0));
+#if GTK_MAJOR_VERSION == 2
+  GtkBox *col = GTK_BOX(gtk_hbox_new(FALSE, 0));
+#else
+  GtkBox *col = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
+#endif
+
   GtkEventBox *row = GTK_EVENT_BOX(gtk_event_box_new());
 
   GtkLabel *show_command     = GTK_LABEL(gtk_label_new(NULL));
   GtkLabel *show_description = GTK_LABEL(gtk_label_new(NULL));
 
   gtk_misc_set_alignment(GTK_MISC(show_command),     0.0, 0.0);
-  gtk_misc_set_alignment(GTK_MISC(show_description), 0.0, 0.0);
+  gtk_misc_set_alignment(GTK_MISC(show_description), 1.0, 0.0);
 
   if (group == true) {
     gtk_misc_set_padding(GTK_MISC(show_command),     2, 4);
@@ -520,6 +511,9 @@ girara_completion_row_create(girara_session_t* session, const char* command, con
   gtk_label_set_use_markup(show_command,     TRUE);
   gtk_label_set_use_markup(show_description, TRUE);
 
+  gtk_label_set_ellipsize(show_command, PANGO_ELLIPSIZE_END);
+  gtk_label_set_ellipsize(show_description, PANGO_ELLIPSIZE_END);
+
   gchar* c = g_markup_printf_escaped(FORMAT_COMMAND,     command ? command : "");
   gchar* d = g_markup_printf_escaped(FORMAT_DESCRIPTION, description ? description : "");
   gtk_label_set_markup(show_command,     c);
@@ -541,7 +535,7 @@ girara_completion_row_create(girara_session_t* session, const char* command, con
   gtk_widget_override_font(GTK_WIDGET(show_description), session->style.font);
 
   gtk_box_pack_start(GTK_BOX(col), GTK_WIDGET(show_command),     TRUE,  TRUE,  2);
-  gtk_box_pack_start(GTK_BOX(col), GTK_WIDGET(show_description), FALSE, FALSE, 2);
+  gtk_box_pack_start(GTK_BOX(col), GTK_WIDGET(show_description), TRUE, TRUE, 2);
 
   gtk_container_add(GTK_CONTAINER(row), GTK_WIDGET(col));
   gtk_widget_show_all(GTK_WIDGET(row));
