@@ -12,63 +12,33 @@
 #include "shortcuts.h"
 #include "config.h"
 #include "utils.h"
+#include "input-history.h"
 
 #if GTK_MAJOR_VERSION == 2
 #include "gtk2-compat.h"
 #endif
 
-static void
-cb_window_icon(girara_session_t* session, const char* UNUSED(name), girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
-{
-  g_return_if_fail(session != NULL && value != NULL);
-
-  if (session->gtk.window != NULL) {
-    GError* error = NULL;
-    gtk_window_set_icon_from_file(GTK_WINDOW(session->gtk.window), (const char*) value, &error);
-    if (error != NULL) {
-      girara_error("failed to load window icon: %s", error->message);
-      g_error_free(error);
-    }
-  }
-}
-
-static void
-cb_font(girara_session_t* session, const char* UNUSED(name), girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
-{
-  g_return_if_fail(session != NULL && value != NULL);
-
-  pango_font_description_free(session->style.font);
-
-  /* parse font */
-  PangoFontDescription* font = pango_font_description_from_string(value);
-  session->style.font = font;
-
-  /* inputbar */
-  if (session->gtk.inputbar_entry != NULL) {
-    gtk_widget_override_font(GTK_WIDGET(session->gtk.inputbar_entry), font);
-  }
-
-  if (session->gtk.inputbar_dialog != NULL) {
-    gtk_widget_override_font(GTK_WIDGET(session->gtk.inputbar_dialog), font);
-  }
-
-  /* notification area */
-  if (session->gtk.notification_text != NULL) {
-    gtk_widget_override_font(GTK_WIDGET(session->gtk.notification_text), font);
-  }
-}
-
 static int
 cb_sort_settings(girara_setting_t* lhs, girara_setting_t* rhs)
 {
   return g_strcmp0(girara_setting_get_name(lhs), girara_setting_get_name(rhs));
 }
 
+static void
+ensure_gettext_initialized(void)
+{
+  static gsize initialized = 0;
+  if (g_once_init_enter(&initialized) == TRUE) {
+    bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
+    bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
+    g_once_init_leave(&initialized, 1);
+  }
+}
+
 girara_session_t*
 girara_session_create()
 {
-  bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
-  bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
+  ensure_gettext_initialized();
 
   girara_session_t* session = g_slice_alloc0(sizeof(girara_session_t));
 
@@ -83,18 +53,24 @@ girara_session_create()
       (girara_free_function_t) girara_shortcut_free);
   session->bindings.inputbar_shortcuts = girara_list_new2(
       (girara_free_function_t) girara_inputbar_shortcut_free);
-  session->elements.statusbar_items    = girara_list_new2(
+
+  session->elements.statusbar_items = girara_list_new2(
       (girara_free_function_t) girara_statusbar_item_free);
-  session->settings                    = girara_sorted_list_new2(
+
+  session->settings = girara_sorted_list_new2(
       (girara_compare_function_t) cb_sort_settings,
       (girara_free_function_t) girara_setting_free);
 
+  /* init modes */
   session->modes.identifiers  = girara_list_new2(
-      (girara_free_function_t) girara_mode_string_free);
+  (girara_free_function_t) girara_mode_string_free);
   girara_mode_t normal_mode   = girara_mode_add(session, "normal");
+  girara_mode_t inputbar_mode = girara_mode_add(session, "inputbar");
   session->modes.normal       = normal_mode;
   session->modes.current_mode = normal_mode;
+  session->modes.inputbar     = inputbar_mode;
 
+  /* config handles */
   session->config.handles           = girara_list_new2(
       (girara_free_function_t) girara_config_handle_free);
   session->config.shortcut_mappings = girara_list_new2(
@@ -102,92 +78,36 @@ girara_session_create()
   session->config.argument_mappings = girara_list_new2(
       (girara_free_function_t) girara_argument_mapping_free);
 
-  session->style.font = NULL;
+  /* command history */
+  session->command_history = girara_input_history_new(NULL);
+  session->global.command_history = girara_get_command_history(session);
 
-  /* default values */
-  int window_width       = 800;
-  int window_height      = 600;
-  int n_completion_items = 15;
-  bool show_scrollbars   = false;
+  /* load default values */
+  girara_config_load_default(session);
 
-  /* add default settings */
-  girara_setting_add(session, "font",                     "monospace normal 9", STRING,  FALSE, _("Font"), cb_font, NULL);
-  girara_setting_add(session, "default-fg",               "#DDDDDD",            STRING,  TRUE,  _("Default foreground color"), NULL, NULL);
-  girara_setting_add(session, "default-bg",               "#000000",            STRING,  TRUE,  _("Default background color"), NULL, NULL);
-  girara_setting_add(session, "inputbar-fg",              "#9FBC00",            STRING,  TRUE,  _("Inputbar foreground color"), NULL, NULL);
-  girara_setting_add(session, "inputbar-bg",              "#131313",            STRING,  TRUE,  _("Inputbar background color"), NULL, NULL);
-  girara_setting_add(session, "statusbar-fg",             "#FFFFFF",            STRING,  TRUE,  _("Statusbar foreground color"), NULL, NULL);
-  girara_setting_add(session, "statusbar-bg",             "#000000",            STRING,  TRUE,  _("Statsubar background color"), NULL, NULL);
-  girara_setting_add(session, "completion-fg",            "#DDDDDD",            STRING,  TRUE,  _("Completion foeground color"), NULL, NULL);
-  girara_setting_add(session, "completion-bg",            "#232323",            STRING,  TRUE,  _("Completion background color"), NULL, NULL);
-  girara_setting_add(session, "completion-group-fg",      "#DEDEDE",            STRING,  TRUE,  _("Completion group foreground color"), NULL, NULL);
-  girara_setting_add(session, "completion-group-bg",      "#000000",            STRING,  TRUE,  _("Completion group background color"), NULL, NULL);
-  girara_setting_add(session, "completion-highlight-fg",  "#232323",            STRING,  TRUE,  _("Completion highlight foreground color"), NULL, NULL);
-  girara_setting_add(session, "completion-highlight-bg",  "#9FBC00",            STRING,  TRUE,  _("Completion highlight background color"), NULL, NULL);
-  girara_setting_add(session, "notification-error-fg",    "#FFFFFF",            STRING,  TRUE,  _("Error notification foreground color"), NULL, NULL);
-  girara_setting_add(session, "notification-error-bg",    "#FF1212",            STRING,  TRUE,  _("Error notification background color"), NULL, NULL);
-  girara_setting_add(session, "notification-warning-fg",  "#000000",            STRING,  TRUE,  _("Warning notification foreground color"), NULL, NULL);
-  girara_setting_add(session, "notification-warning-bg",  "#F3F000",            STRING,  TRUE,  _("Warning notifaction background color"), NULL, NULL);
-  girara_setting_add(session, "notification-fg",          "#000000",            STRING,  TRUE,  _("Notification foreground color"), NULL, NULL);
-  girara_setting_add(session, "notification-bg",          "#FFFFFF",            STRING,  TRUE,  _("Notification background color"), NULL, NULL);
-  girara_setting_add(session, "tabbar-fg",                "#939393",            STRING,  TRUE,  _("Tab bar foreground color"), NULL, NULL);
-  girara_setting_add(session, "tabbar-bg",                "#000000",            STRING,  TRUE,  _("Tab bar background color"), NULL, NULL);
-  girara_setting_add(session, "tabbar-focus-fg",          "#9FBC00",            STRING,  TRUE,  _("Tab bar foreground color (active)"), NULL, NULL);
-  girara_setting_add(session, "tabbar-focus-bg",          "#000000",            STRING,  TRUE,  _("Tab bar background color (active)"), NULL, NULL);
-  girara_setting_add(session, "word-separator",           " /.-=&#?",           STRING,  TRUE,  NULL, NULL, NULL);
-  girara_setting_add(session, "window-width",             &window_width,        INT,     TRUE,  _("Initial window width"), NULL, NULL);
-  girara_setting_add(session, "window-height",            &window_height,       INT,     TRUE,  _("Initial window height"), NULL, NULL);
-  girara_setting_add(session, "n-completion-items",       &n_completion_items,  INT,     TRUE,  _("Number of completion items"), NULL, NULL);
-  girara_setting_add(session, "show-scrollbars",          &show_scrollbars,     BOOLEAN, TRUE,  _("Show scrollbars"), NULL, NULL);
-
-  /* window icon */
-  girara_setting_add(session, "window-icon",              "",                   STRING,  FALSE, _("Window icon"), cb_window_icon, NULL);
-  /* exec-command */
-  girara_setting_add(session, "exec-command",             "",                   STRING,  FALSE, _("Command to execute in :exec"), NULL, NULL);
-
-  /* default shortcuts */
-  girara_shortcut_add(session, 0,                GDK_KEY_Escape, NULL, girara_sc_abort,           normal_mode, 0, NULL);
-  girara_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_c,      NULL, girara_sc_abort,           normal_mode, 0, NULL);
-  girara_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_q,      NULL, girara_sc_quit,            normal_mode, 0, NULL);
-  girara_shortcut_add(session, 0,                GDK_KEY_colon,  NULL, girara_sc_focus_inputbar,  normal_mode, 0, ":");
-  girara_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_w,      NULL, girara_sc_tab_close,       normal_mode, 0, NULL);
-  girara_shortcut_add(session, 0,                0,              "gt", girara_sc_tab_navigate,    normal_mode, GIRARA_NEXT,     NULL);
-  girara_shortcut_add(session, 0,                0,              "gT", girara_sc_tab_navigate,    normal_mode, GIRARA_PREVIOUS, NULL);
-
-  /* default inputbar shortcuts */
-  girara_inputbar_shortcut_add(session, 0,                GDK_KEY_Escape,       girara_isc_abort,               0,                           NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_c,            girara_isc_abort,               0,                           NULL);
-  girara_inputbar_shortcut_add(session, 0,                GDK_KEY_Tab,          girara_isc_completion,          GIRARA_NEXT,                 NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_Tab,          girara_isc_completion,          GIRARA_NEXT_GROUP,           NULL);
-  girara_inputbar_shortcut_add(session, 0,                GDK_KEY_ISO_Left_Tab, girara_isc_completion,          GIRARA_PREVIOUS,             NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_ISO_Left_Tab, girara_isc_completion,          GIRARA_PREVIOUS_GROUP,       NULL);
-  girara_inputbar_shortcut_add(session, 0,                GDK_KEY_BackSpace,    girara_isc_string_manipulation, GIRARA_DELETE_LAST_CHAR,     NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_h,            girara_isc_string_manipulation, GIRARA_DELETE_LAST_CHAR,     NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_u,            girara_isc_string_manipulation, GIRARA_DELETE_TO_LINE_START, NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_k,            girara_isc_string_manipulation, GIRARA_DELETE_TO_LINE_END,   NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_d,            girara_isc_string_manipulation, GIRARA_DELETE_CURR_CHAR,     NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_w,            girara_isc_string_manipulation, GIRARA_DELETE_LAST_WORD,     NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_f,            girara_isc_string_manipulation, GIRARA_NEXT_CHAR,            NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_b,            girara_isc_string_manipulation, GIRARA_PREVIOUS_CHAR,        NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_a,            girara_isc_string_manipulation, GIRARA_GOTO_START,           NULL);
-  girara_inputbar_shortcut_add(session, GDK_CONTROL_MASK, GDK_KEY_e,            girara_isc_string_manipulation, GIRARA_GOTO_END,             NULL);
-
-  /* default commands */
-  girara_inputbar_command_add(session, "exec",  NULL, girara_cmd_exec,  NULL,          _("Execute a command"));
-  girara_inputbar_command_add(session, "map",   "m",  girara_cmd_map,   NULL,          _("Map a key sequence"));
-  girara_inputbar_command_add(session, "quit",  "q",  girara_cmd_quit,  NULL,          _("Quit the program"));
-  girara_inputbar_command_add(session, "set",   "s",  girara_cmd_set,   girara_cc_set, _("Set an option"));
-  girara_inputbar_command_add(session, "unmap", NULL, girara_cmd_unmap, NULL,          _("Unmap a key sequence"));
-
-  /* default config handle */
-  girara_config_handle_add(session, "map",   girara_cmd_map);
-  girara_config_handle_add(session, "set",   girara_cmd_set);
-  girara_config_handle_add(session, "unmap", girara_cmd_unmap);
-
-  /* default shortcut mappings */
-  girara_shortcut_mapping_add(session, "focus_inputbar", girara_sc_focus_inputbar);
-  girara_shortcut_mapping_add(session, "quit",           girara_sc_quit);
-  girara_shortcut_mapping_add(session, "set",            girara_sc_set);
+  /* create widgets */
+#if GTK_MAJOR_VERSION == 2
+  session->gtk.box               = GTK_BOX(gtk_vbox_new(FALSE, 0));
+  session->gtk.statusbar_entries = GTK_BOX(gtk_hbox_new(FALSE, 0));
+  session->gtk.tabbar            = gtk_hbox_new(TRUE, 0);
+  session->gtk.inputbar_box      = GTK_BOX(gtk_hbox_new(TRUE, 0));
+#else
+  session->gtk.box               = GTK_BOX(gtk_box_new(GTK_ORIENTATION_VERTICAL, 0));
+  session->gtk.statusbar_entries = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
+  session->gtk.tabbar            = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
+  session->gtk.inputbar_box      = GTK_BOX(gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0));
+  gtk_box_set_homogeneous(GTK_BOX(session->gtk.tabbar), TRUE);
+  gtk_box_set_homogeneous(session->gtk.inputbar_box, TRUE);
+#endif
+  session->gtk.view              = gtk_scrolled_window_new(NULL, NULL);
+  session->gtk.viewport          = gtk_viewport_new(NULL, NULL);
+  session->gtk.statusbar         = gtk_event_box_new();
+  session->gtk.notification_area = gtk_event_box_new();
+  session->gtk.notification_text = gtk_label_new(NULL);
+  session->gtk.inputbar_dialog   = GTK_LABEL(gtk_label_new(NULL));
+  session->gtk.inputbar_entry    = GTK_ENTRY(gtk_entry_new());
+  session->gtk.inputbar          = gtk_event_box_new();
+  session->gtk.tabs              = GTK_NOTEBOOK(gtk_notebook_new());
 
   return session;
 }
@@ -195,30 +115,21 @@ girara_session_create()
 bool
 girara_session_init(girara_session_t* session, const char* sessionname)
 {
+  if (session == NULL) {
+    return false;
+  }
+
+  /* window */
   if (session->gtk.embed){
     session->gtk.window = gtk_plug_new(session->gtk.embed);
   } else {
     session->gtk.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   }
+
   if (sessionname != NULL) {
     gtk_widget_set_name(GTK_WIDGET(session->gtk.window), sessionname);
   }
 
-  session->gtk.box               = GTK_BOX(gtk_vbox_new(FALSE, 0));
-  session->gtk.view              = gtk_scrolled_window_new(NULL, NULL);
-  session->gtk.viewport          = gtk_viewport_new(NULL, NULL);
-  session->gtk.statusbar         = gtk_event_box_new();
-  session->gtk.statusbar_entries = GTK_BOX(gtk_hbox_new(FALSE, 0));
-  session->gtk.notification_area = gtk_event_box_new();
-  session->gtk.notification_text = gtk_label_new(NULL);
-  session->gtk.tabbar            = gtk_hbox_new(TRUE, 0);
-  session->gtk.inputbar_box      = GTK_BOX(gtk_hbox_new(TRUE, 0));
-  session->gtk.inputbar_dialog   = GTK_LABEL(gtk_label_new(NULL));
-  session->gtk.inputbar_entry    = GTK_ENTRY(gtk_entry_new());
-  session->gtk.inputbar          = gtk_event_box_new();
-  session->gtk.tabs              = GTK_NOTEBOOK(gtk_notebook_new());
-
-  /* window */
   GdkGeometry hints = {
     .base_height = 1,
     .base_width  = 1,
@@ -238,6 +149,12 @@ girara_session_init(girara_session_t* session, const char* sessionname)
   gtk_window_set_has_resize_grip(GTK_WINDOW(session->gtk.window), FALSE);
 #endif
 
+#if (GTK_MAJOR_VERSION == 3)
+  gtk_widget_override_background_color(GTK_WIDGET(session->gtk.window), GTK_STATE_NORMAL, &(session->style.default_background));
+#else
+  gtk_widget_modify_bg(GTK_WIDGET(session->gtk.window), GTK_STATE_NORMAL, &(session->style.default_background));
+#endif
+
   /* view */
   session->signals.view_key_pressed = g_signal_connect(G_OBJECT(session->gtk.view), "key-press-event",
       G_CALLBACK(girara_callback_view_key_press_event), session);
@@ -251,7 +168,7 @@ girara_session_init(girara_session_t* session, const char* sessionname)
   session->signals.view_motion_notify_event = g_signal_connect(G_OBJECT(session->gtk.view), "motion-notify-event",
       G_CALLBACK(girara_callback_view_button_motion_notify_event), session);
 
-  session->signals.view_key_pressed = g_signal_connect(G_OBJECT(session->gtk.view), "scroll-event",
+  session->signals.view_scroll_event = g_signal_connect(G_OBJECT(session->gtk.view), "scroll-event",
       G_CALLBACK(girara_callback_view_scroll_event), session);
 
   bool tmp_bool_value = false;
@@ -268,6 +185,12 @@ girara_session_init(girara_session_t* session, const char* sessionname)
   gtk_container_add(GTK_CONTAINER(session->gtk.view), session->gtk.viewport);
   gtk_viewport_set_shadow_type(GTK_VIEWPORT(session->gtk.viewport), GTK_SHADOW_NONE);
 
+#if (GTK_MAJOR_VERSION == 3)
+  gtk_widget_override_background_color(GTK_WIDGET(session->gtk.viewport), GTK_STATE_NORMAL, &(session->style.default_background));
+#else
+  gtk_widget_modify_bg(GTK_WIDGET(session->gtk.viewport), GTK_STATE_NORMAL, &(session->style.default_background));
+#endif
+
   /* box */
   gtk_box_set_spacing(session->gtk.box, 0);
   gtk_container_add(GTK_CONTAINER(session->gtk.window), GTK_WIDGET(session->gtk.box));
@@ -427,9 +350,16 @@ girara_session_init(girara_session_t* session, const char* sessionname)
 
   gtk_widget_show_all(GTK_WIDGET(session->gtk.window));
   gtk_widget_hide(GTK_WIDGET(session->gtk.notification_area));
-  gtk_widget_hide(GTK_WIDGET(session->gtk.inputbar));
   gtk_widget_hide(GTK_WIDGET(session->gtk.inputbar_dialog));
 
+  if (session->global.autohide_inputbar == true) {
+    gtk_widget_hide(GTK_WIDGET(session->gtk.inputbar));
+  }
+
+  if (session->global.hide_statusbar == true) {
+    gtk_widget_hide(GTK_WIDGET(session->gtk.statusbar));
+  }
+
   char* window_icon = NULL;
   girara_setting_get(session, "window-icon", &window_icon);
   if (window_icon != NULL) {
@@ -439,7 +369,9 @@ girara_session_init(girara_session_t* session, const char* sessionname)
     g_free(window_icon);
   }
 
-  return TRUE;
+  gtk_widget_grab_focus(GTK_WIDGET(session->gtk.view));
+
+  return true;
 }
 
 bool
@@ -448,7 +380,9 @@ girara_session_destroy(girara_session_t* session)
   g_return_val_if_fail(session != NULL, FALSE);
 
   /* clean up style */
-  pango_font_description_free(session->style.font);
+  if (session->style.font) {
+    pango_font_description_free(session->style.font);
+  }
 
   /* clean up shortcuts */
   girara_list_free(session->bindings.shortcuts);
@@ -474,6 +408,10 @@ girara_session_destroy(girara_session_t* session)
   girara_list_free(session->settings);
   session->settings = NULL;
 
+  /* clean up input histry */
+  g_object_unref(session->command_history);
+  session->command_history = NULL;
+
   /* clean up statusbar items */
   girara_list_free(session->elements.statusbar_items);
   session->elements.statusbar_items = NULL;
@@ -682,3 +620,9 @@ girara_set_window_title(girara_session_t* session, const char* name)
   return true;
 }
 
+girara_list_t*
+girara_get_command_history(girara_session_t* session)
+{
+  g_return_val_if_fail(session != NULL, FALSE);
+  return girara_input_history_list(session->command_history);
+}
