@@ -6,11 +6,13 @@
 #include "session.h"
 #include "settings.h"
 #include "tabs.h"
+#include "input-history.h"
 
 #include <string.h>
 #include <gtk/gtk.h>
 
 static void girara_toggle_widget_visibility(GtkWidget* widget);
+static bool simulate_key_press(girara_session_t* session, int state, int key);
 
 bool
 girara_shortcut_add(girara_session_t* session, guint modifier, guint key, const char* buffer, girara_shortcut_function_t function, girara_mode_t mode, int argument_n, void* argument_data)
@@ -144,9 +146,12 @@ girara_inputbar_shortcut_free(girara_inputbar_shortcut_t* inputbar_shortcut)
 {
   g_slice_free(girara_inputbar_shortcut_t, inputbar_shortcut);
 }
+
 bool
 girara_isc_abort(girara_session_t* session, girara_argument_t* UNUSED(argument), girara_event_t* UNUSED(event), unsigned int UNUSED(t))
 {
+  g_return_val_if_fail(session != NULL, false);
+
   /* hide completion */
   girara_argument_t arg = { GIRARA_HIDE, NULL };
   girara_isc_completion(session, &arg, NULL, 0);
@@ -159,7 +164,12 @@ girara_isc_abort(girara_session_t* session, girara_argument_t* UNUSED(argument),
 
   /* hide inputbar */
   gtk_widget_hide(GTK_WIDGET(session->gtk.inputbar_dialog));
-  gtk_widget_hide(GTK_WIDGET(session->gtk.inputbar));
+  if (session->global.autohide_inputbar == true) {
+    gtk_widget_hide(GTK_WIDGET(session->gtk.inputbar));
+  }
+
+  /* Begin from the last command when navigating through history */
+  girara_input_history_reset(session->command_history);
 
   /* reset custom functions */
   session->signals.inputbar_custom_activate        = NULL;
@@ -172,6 +182,8 @@ girara_isc_abort(girara_session_t* session, girara_argument_t* UNUSED(argument),
 bool
 girara_isc_string_manipulation(girara_session_t* session, girara_argument_t* argument, girara_event_t* UNUSED(event), unsigned int UNUSED(t))
 {
+  g_return_val_if_fail(session != NULL, false);
+
   gchar *separator = NULL;
   girara_setting_get(session, "word-separator", &separator);
   gchar *input  = gtk_editable_get_chars(GTK_EDITABLE(session->gtk.inputbar_entry), 0, -1);
@@ -181,11 +193,14 @@ girara_isc_string_manipulation(girara_session_t* session, girara_argument_t* arg
 
   switch (argument->n) {
     case GIRARA_DELETE_LAST_WORD:
-      i = pos - 1;
-
-      if (!pos) {
+      if (pos == 1 && (input[0] == ':' || input[0] == '/')) {
         break;
       }
+      if (pos == 0) {
+	      break;
+      }
+
+      i = pos - 1;
 
       /* remove trailing spaces */
       for (; i >= 0 && input[i] == ' '; i--);
@@ -199,7 +214,10 @@ girara_isc_string_manipulation(girara_session_t* session, girara_argument_t* arg
       gtk_editable_set_position(GTK_EDITABLE(session->gtk.inputbar_entry), i + 1);
       break;
     case GIRARA_DELETE_LAST_CHAR:
-      if ((length - 1) <= 0) {
+      if (length != 1 && pos == 1 && (input[0] == ':' || input[0] == '/')) {
+        break;
+      }
+      if (length == 1 && pos == 1) {
         girara_isc_abort(session, argument, NULL, 0);
       }
       gtk_editable_delete_text(GTK_EDITABLE(session->gtk.inputbar_entry), pos - 1, pos);
@@ -211,10 +229,13 @@ girara_isc_string_manipulation(girara_session_t* session, girara_argument_t* arg
       gtk_editable_set_position(GTK_EDITABLE(session->gtk.inputbar_entry), pos + 1);
       break;
     case GIRARA_PREVIOUS_CHAR:
-      gtk_editable_set_position(GTK_EDITABLE(session->gtk.inputbar_entry), (pos == 0) ? 0 : pos - 1);
+      gtk_editable_set_position(GTK_EDITABLE(session->gtk.inputbar_entry), (pos == 1) ? 1 : pos - 1);
       break;
     case GIRARA_DELETE_CURR_CHAR:
-      if((length - 1) <= 0) {
+      if (length != 1 && pos == 0 && (input[0] == ':' || input[0] == '/')){
+	break;
+      }
+      if(length == 1 && pos == 0) {
         girara_isc_abort(session, argument, NULL, 0);
       }
       gtk_editable_delete_text(GTK_EDITABLE(session->gtk.inputbar_entry), pos, pos + 1);
@@ -236,6 +257,27 @@ girara_isc_string_manipulation(girara_session_t* session, girara_argument_t* arg
   return false;
 }
 
+bool
+girara_isc_command_history(girara_session_t* session, girara_argument_t*
+    argument, girara_event_t* UNUSED(event), unsigned int UNUSED(t))
+{
+  g_return_val_if_fail(session != NULL, false);
+
+  char* temp = gtk_editable_get_chars(GTK_EDITABLE(session->gtk.inputbar_entry), 0, -1);
+  const char* command = argument->n == GIRARA_NEXT ?
+    girara_input_history_next(session->command_history, temp) :
+    girara_input_history_previous(session->command_history, temp);
+  g_free(temp);
+
+  if (command != NULL) {
+    gtk_entry_set_text(session->gtk.inputbar_entry, command);
+    gtk_widget_grab_focus(GTK_WIDGET(session->gtk.inputbar_entry));
+    gtk_editable_set_position(GTK_EDITABLE(session->gtk.inputbar_entry), -1);
+  }
+
+  return true;
+}
+
 /* default shortcut implementation */
 bool
 girara_sc_focus_inputbar(girara_session_t* session, girara_argument_t* argument, girara_event_t* UNUSED(event), unsigned int UNUSED(t))
@@ -277,8 +319,13 @@ girara_sc_abort(girara_session_t* session, girara_argument_t* UNUSED(argument),
   g_return_val_if_fail(session != NULL, false);
 
   girara_isc_abort(session, NULL, NULL, 0);
+
   gtk_widget_hide(GTK_WIDGET(session->gtk.notification_area));
 
+  if (session->global.autohide_inputbar == false) {
+    gtk_widget_show(GTK_WIDGET(session->gtk.inputbar));
+  }
+
   return false;
 }
 
@@ -315,6 +362,9 @@ girara_sc_tab_navigate(girara_session_t* session, girara_argument_t* argument, g
   g_return_val_if_fail(session != NULL, false);
 
   unsigned int number_of_tabs = girara_get_number_of_tabs(session);
+  if (number_of_tabs == 0)
+    return false;
+
   unsigned int current_tab    = girara_tab_position_get(session, girara_tab_current_get(session));
   unsigned int step           = (argument->n == GIRARA_PREVIOUS) ? -1 : 1;
   unsigned int new_tab        = (current_tab + step) % number_of_tabs;
@@ -416,7 +466,129 @@ girara_sc_set(girara_session_t* session, girara_argument_t* argument, girara_eve
   return false;
 }
 
-bool girara_shortcut_mapping_add(girara_session_t* session, const char* identifier, girara_shortcut_function_t function)
+bool
+girara_sc_feedkeys(girara_session_t* session, girara_argument_t* argument,
+    girara_event_t* UNUSED(event), unsigned int t)
+{
+  if (session == NULL || argument == NULL) {
+    return false;
+  }
+
+  typedef struct gdk_keyboard_button_s {
+    char* identifier;
+    int keyval;
+  } gdk_keyboard_button_t;
+
+  static const gdk_keyboard_button_t gdk_keyboard_buttons[] = {
+    {"BackSpace", GDK_KEY_BackSpace},
+    {"CapsLock",  GDK_KEY_Caps_Lock},
+    {"Down",      GDK_KEY_Down},
+    {"Esc",       GDK_KEY_Escape},
+    {"F10",       GDK_KEY_F10},
+    {"F11",       GDK_KEY_F11},
+    {"F12",       GDK_KEY_F12},
+    {"F1",        GDK_KEY_F1},
+    {"F2",        GDK_KEY_F2},
+    {"F3",        GDK_KEY_F3},
+    {"F4",        GDK_KEY_F4},
+    {"F5",        GDK_KEY_F5},
+    {"F6",        GDK_KEY_F6},
+    {"F7",        GDK_KEY_F7},
+    {"F8",        GDK_KEY_F8},
+    {"F9",        GDK_KEY_F9},
+    {"Left",      GDK_KEY_Left},
+    {"PageDown",  GDK_KEY_Page_Down},
+    {"PageUp",    GDK_KEY_Page_Up},
+    {"Home",      GDK_KEY_Home},
+    {"End",       GDK_KEY_End},
+    {"Return",    GDK_KEY_Return},
+    {"Right",     GDK_KEY_Right},
+    {"Space",     GDK_KEY_space},
+    {"Super",     GDK_KEY_Super_L},
+    {"Tab",       GDK_KEY_Tab},
+    {"ShiftTab",  GDK_KEY_ISO_Left_Tab},
+    {"Up",        GDK_KEY_Up}
+  };
+
+  char* input               = (char*) argument->data;
+  unsigned int input_length = strlen(input);
+
+  t = (t == 0) ? 1 : t;
+  for (unsigned int c = 0; c < t; c++) {
+    for (unsigned i = 0; i < input_length; i++) {
+      int state  = 0;
+      int keyval = input[i];
+
+      /* possible special button */
+      if ((input_length - i) >= 3 && input[i] == '<') {
+        char* end = strchr(input + i, '>');
+        if (end == NULL) {
+          goto single_key;
+        }
+
+        int length = end - (input + i) - 1;
+        char* tmp  = g_strndup(input + i + 1, length);
+        bool found = false;
+
+        /* Multi key shortcut */
+        if (length > 2 && tmp[1] == '-') {
+          switch (tmp[0]) {
+            case 'S':
+              state = GDK_SHIFT_MASK;
+              break;
+            case 'A':
+              state = GDK_MOD1_MASK;
+              break;
+            case 'C':
+              state = GDK_CONTROL_MASK;
+              break;
+            default:
+              break;
+          }
+
+          if (length == 3) {
+            keyval = tmp[2];
+            found  = true;
+          } else {
+            for (unsigned int i = 0; i < LENGTH(gdk_keyboard_buttons); i++) {
+              if (g_strcmp0(tmp + 2, gdk_keyboard_buttons[i].identifier) == 0) {
+                keyval = gdk_keyboard_buttons[i].keyval;
+                found = true;
+                break;
+              }
+            }
+          }
+        /* Possible special key */
+        } else {
+          for (unsigned int i = 0; i < LENGTH(gdk_keyboard_buttons); i++) {
+            if (g_strcmp0(tmp, gdk_keyboard_buttons[i].identifier) == 0) {
+              keyval = gdk_keyboard_buttons[i].keyval;
+              found = true;
+              break;
+            }
+          }
+        }
+
+        g_free(tmp);
+
+        /* parsed special key */
+        if (found == true) {
+          i += length + 1;
+        }
+      }
+
+single_key:
+
+      update_state_by_keyval(&state, keyval);
+      simulate_key_press(session, state, keyval);
+    }
+  }
+
+  return true;
+}
+
+bool
+girara_shortcut_mapping_add(girara_session_t* session, const char* identifier, girara_shortcut_function_t function)
 {
   g_return_val_if_fail(session  != NULL, false);
 
@@ -453,7 +625,8 @@ girara_shortcut_mapping_free(girara_shortcut_mapping_t* mapping)
   g_slice_free(girara_shortcut_mapping_t, mapping);
 }
 
-bool girara_argument_mapping_add(girara_session_t* session, const char* identifier, int value)
+bool
+girara_argument_mapping_add(girara_session_t* session, const char* identifier, int value)
 {
   g_return_val_if_fail(session  != NULL, false);
 
@@ -553,3 +726,43 @@ girara_mouse_event_free(girara_mouse_event_t* mouse_event)
   }
   g_slice_free(girara_mouse_event_t, mouse_event);
 }
+
+static bool
+simulate_key_press(girara_session_t* session, int state, int key)
+{
+  if (session == NULL || session->gtk.box == NULL) {
+    return false;
+  }
+
+  GdkEvent* event = gdk_event_new(GDK_KEY_PRESS);
+
+  event->key.type       = GDK_KEY_PRESS;
+  event->key.window     = gtk_widget_get_parent_window(GTK_WIDGET(session->gtk.box));
+  event->key.send_event = false;
+  event->key.time       = GDK_CURRENT_TIME;
+  event->key.state      = state;
+  event->key.keyval     = key;
+
+  g_object_ref(event->key.window);
+
+  GdkKeymapKey* keys;
+  gint number_of_keys;
+
+  if (gdk_keymap_get_entries_for_keyval(gdk_keymap_get_default(),
+        event->key.keyval, &keys, &number_of_keys) == FALSE) {
+    gdk_event_free(event);
+    return false;
+  }
+
+  event->key.hardware_keycode = keys[0].keycode;
+  event->key.group            = keys[0].group;
+
+  g_free(keys);
+
+  gdk_event_put(event);
+  gdk_event_free(event);
+
+  gtk_main_iteration_do(FALSE);
+
+  return true;
+}
